function NK2() {}
NK2.debug=3;//set debug level
NK2.LEDflasheson=true;//false disables led flash timers for debugging purposes

//############################################################################
//defaults
//############################################################################

NK2.midiChannel=0xB0;
NK2.curNSMR="N";//default bank (NSMR refers to SMR buttons and "no button")
NK2.curDeck=0;//current deck (for bank selects) -- defaults to 0 = off - no deck selected
NK2.bankSelectState=0;//current state of marker set button (deck selector button - press marker set, hold, press S M or R button on any track to select that bank) -- future: marker set could also select left buttons, transport, etc for more banks
NK2.cycleState=0;//current state of cycle toggle button
NK2.MOD1state=0;
NK2.MOD2state=0;
NK2.LEDBankIndicator=false;
NK2.loopmove=0.0125;//how far to move loops with loop move functions
NK2.lastwavevalue=0;//used for wave zoom function

//############################################################################
//references
//############################################################################
 
NK2.MODcodes={"0000":0,"0001":1,"0010":2,"0011":3,"0100":4,"0101":5,"0110":6,"0111":7,"1000":8,"1001":9,"1010":10,"1011":11,"1100":12,"1101":13,"1110":14,"1111":15};//list of mod states
NK2.Banks=new Array("N", "S", "M", "R");

//list controls
NK2.Knob={1:0x10,2:0x11,3:0x12,4:0x13,5:0x14,6:0x15,7:0x16,8:0x17};
NK2.Sbutton={1:0x20,2:0x21,3:0x22,4:0x23,5:0x24,6:0x25,7:0x26,8:0x27};
NK2.Mbutton={1:0x30,2:0x31,3:0x32,4:0x33,5:0x34,6:0x35,7:0x36,8:0x37};
NK2.Rbutton={1:0x40,2:0x41,3:0x42,4:0x43,5:0x44,6:0x45,7:0x46,8:0x47};
NK2.leftButton={"trdown":0x3A,"trup":0x3B,"cycle":0x2E,"mset":0x3C,"mdown":0x3D,"mup":0x3E,"rev":0x2B,"ff":0x2C,"stop":0x2A,"play":0x29,"rec":0x2D};

NK2.faders={0x00:1,0x01:2,0x02:3,0x03:4,0x04:5,0x05:6,0x06:7,0x07:8};

NK2.Deck={1:"[Channel1]",2:"[Channel2]",3:"[Channel3]",4:"[Channel4]",5:"[Sampler1]",6:"[Sampler2]",7:"[Sampler3]",8:"[Sampler4]"};//list of decks, applied to each strip - 8 strips, 8 decks.  (4 decks, 4 samplers)
NK2.DeckNum={"[Channel1]":1,"[Channel2]":2,"[Channel3]":3,"[Channel4]":4,"[Sampler1]":5,"[Sampler2]":6,"[Sampler3]":7,"[Sampler4]":8};//reverse lookup for deck numbers by name
NK2.deckMuteVol={"[Channel1]":-1,"[Channel2]":-1,"[Channel3]":-1,"[Channel4]":-1,"[Sampler1]":-1,"[Sampler2]":-1,"[Sampler3]":-1,"[Sampler4]":-1};//store volumes of decks for mute function -1 means deck is not muted
	
//############################################################################
//INIT & SHUTDOWN
//############################################################################

NK2.init = function init() { // called when the device is opened & set up
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	engine.setValue("[Master]", "num_decks", 4);
	NK2.setup()
	
	NK2.updateLEDs();
	
	
	};

NK2.shutdown = function shutdown() {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	
	engine.stopTimer(NK2.LEDtimer);
	};
	
//############################################################################
//SCENE SET FUNCTIONS - configure controls for selected scenes
//############################################################################

NK2.bankSelect = function bankSelect(deck, bank) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	NK2.curNSMR=bank;
	NK2.curDeck=deck;
	NK2.doClearBank=false;//bank has just been selected.  Don't clear the bank when bank select button is released
	NK2.cycleState=0;
	NK2.bankSelectState=0;
	
	//find bank indicator led, if bank activated
	if (NK2.curNSMR=="S"){
		NK2.LEDBankIndicator=NK2.Sbutton[NK2.curDeck];
	}else if (NK2.curNSMR=="M"){
		NK2.LEDBankIndicator=NK2.Mbutton[NK2.curDeck];
	}else if (NK2.curNSMR=="R"){
		NK2.LEDBankIndicator=NK2.Rbutton[NK2.curDeck];
	}
	if (NK2.LEDflasheson===true){engine.stopTimer(NK2.LEDtimer);NK2.LEDtimer=engine.beginTimer(750, "NK2.indicatorLEDs()");}//start timer for LED indicator flasher
	NK2.updateLEDs();
};
	
NK2.setMain = function setMain() {//set controls - no bank/track selected
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
};

NK2.setS = function setS() {//set controls - S bank selected
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
};

NK2.setM = function setM() {//set controls - M bank selected
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
};

NK2.setR = function setR() {//set controls - R bank selected
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
};

//############################################################################
//INPUT PROCESSING
//############################################################################


NK2.fader = function fader(channel, control, value, status, group) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	engine.setValue(NK2.Deck[NK2.faders[control]], "volume", value/128);
};

NK2.button = function button(channel, control, value, status, group) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	//input received - gatekeeper function connected to all buttons
 
	//get the control object for the current button on the current bank - associative array containing control info, LED state evaluators, press/release functions, etc.
	var theControl=NK2.getControl(control);
	if (theControl===false){return false;}//no control set for this button
	
	if (value>0){//button was pressed
		var command = theControl["pressEval"];
		midi.sendShortMsg(NK2.midiChannel, control, 0x7F);//light LED while button pressed
	}else{//button was released
		var command = theControl["releaseEval"];
		midi.sendShortMsg(NK2.midiChannel, control, NK2.Controls[control]["LEDstate"]);//restore LED state when button released
	}

	if (command!=false)eval(command);
};

NK2.knob = function knob(channel, control, value, status, group) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	
	//input received - gatekeeper function connected to all knobs
 
	//get the control object for the current button on the current bank - associative array containing control info, LED state evaluators, press/release functions, etc.
	var theControl=NK2.getControl(control);
	if (theControl===false){return false;}//no control set for this button
	
	var command = theControl["pressEval"];
	if (command!=false)eval(command);
};

NK2.cyclepress = function cyclepress() {//press cycle button
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	print("##################START CYCLEPRESS");
	if (NK2.cycleState==0){
		NK2.cycleState=1;
	}else{
		NK2.cycleState=0;
	}
	NK2.updateLEDs();
	print("##################END CYCLEPRESS");
};

NK2.modpress = function modpress(button) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	if(button=="BANK"){
		NK2.doClearBank=true;//if false when released, bank has been switched, so don't clear the selected bank on release
		NK2.bankSelectState=1;
	}else if (button=="MOD1"){
		NK2.MOD1state=1;
	}else if (button=="MOD2"){
		NK2.MOD2state=1;
	};
	NK2.updateLEDs();
};

NK2.modrelease = function modrelease(button) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	if(button=="BANK"){
		NK2.bankSelectState=0;
		if (NK2.doClearBank===true){
			if (NK2.curNSMR!="N"){NK2.cycleState=0;}
			NK2.curNSMR="N";
			NK2.curDeck=0;
			NK2.LEDBankIndicator=false;
			engine.stopTimer(NK2.LEDtimer);
			NK2.updateLEDs();
		}
		return true;
	}else if (button=="MOD1"){
		NK2.MOD1state=0;
	}else if (button=="MOD2"){
		NK2.MOD2state=0;
	};
	NK2.updateLEDs();
};


//############################################################################
//LED FUNCTIONS
//############################################################################

NK2.updateLEDs = function updateLEDs() {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	//fires on mod press/release, bank switch - updates all LEDs
	var midino;
	var control;
	var state=0;
	
	print("----------------UPDATE LEDS-----cycle state: "+NK2.cycleState);
	
	for (var key in NK2.Controls){//clear hooks
		var midino=key;
		var thecontrol = NK2.getControl(key);
		
		
		var LEDstateType=thecontrol["LEDstateType"];
		var LEDstateEval=thecontrol["LEDstateEval"];
		var LEDhookControl=thecontrol["LEDhookControl"];
		var LEDhookFunction=thecontrol["LEDhookFunction"];
		var LEDhookGroup=(thecontrol["LEDhookGroup"]=="default")?NK2.Deck[NK2.curDeck]:thecontrol["LEDhookGroup"];
		
		if(LEDstateType=="hook"){
			print("#######NEW HOOK");
			print("LEDstateType: "+LEDstateType);
			print("LEDstateEval: "+LEDstateEval);
			print("LEDhookControl: "+LEDhookControl);
			print("LEDhookFunction: "+LEDhookFunction);
			print("LEDhookGroup: "+LEDhookGroup);
		}
		
		if(NK2.Controls[key]["LEDstateType"]=="hook"){
			print("#######OLD HOOK");
			print("LEDstateType: "+NK2.Controls[key]["LEDstateType"]);
			print("LEDstateEval: "+NK2.Controls[key]["LEDstateEval"]);
			print("LEDhookControl: "+NK2.Controls[key]["LEDhookControl"]);
			print("LEDhookFunction: "+NK2.Controls[key]["LEDhookFunction"]);
			print("LEDhookGroup: "+NK2.Controls[key]["LEDhookGroup"]);
		}
		
		//clear hooks
		if (NK2.Controls[key]["LEDstateType"]==="hook" && LEDstateType==="hook"){//check if hooks are the same, if not, clear the hook
			if(NK2.Controls[key]["LEDhookGroup"]==LEDhookGroup && NK2.Controls[key]["LEDhookControl"]==LEDhookControl && NK2.Controls[key]["LEDhookFunction"]==LEDhookFunction){
				//hooks are the same - go to next control
				if (NK2.debug>0){print("##################SKIPPING")};
				continue;
			}else{//not the same.  clear the hook
				if (NK2.debug>0){print("##################CLEARING####################")};
				//engine.connectControl(NK2.Controls[key]["LEDhookGroup"], NK2.Controls[key]["LEDhookControl"], NK2.Controls[key]["LEDhookFunction"], true);
				NK2.clearHook(NK2.Controls[key]["LEDhookGroup"], NK2.Controls[key]["LEDhookControl"], NK2.Controls[key]["LEDhookFunction"]);
			}
			
		}else if(NK2.Controls[key]["LEDstateType"]==="hook"){//clear the hook
			//engine.connectControl(NK2.Controls[key]["LEDhookGroup"], NK2.Controls[key]["LEDhookControl"], NK2.Controls[key]["LEDhookFunction"], true);
			NK2.clearHook(NK2.Controls[key]["LEDhookGroup"], NK2.Controls[key]["LEDhookControl"], NK2.Controls[key]["LEDhookFunction"]);
		};
		//engine.connectControl(NK2.Controls[key]["LEDhookGroup"], NK2.Controls[key]["LEDhookControl"], NK2.Controls[key]["LEDhookFunction"], true);
	
		
		if (thecontrol["LEDstateType"]===false){continue;}//control has no LED, LED Type is not set, or LED doesn't change on this mod setting.  skip to next control
		
		//reset LED state storage
		NK2.Controls[key]["LEDstate"]=0;
		NK2.Controls[key]["LEDstateType"]=LEDstateType;
		NK2.Controls[key]["LEDstateEval"]=LEDstateEval;
		NK2.Controls[key]["LEDhookControl"]=LEDhookControl;
		NK2.Controls[key]["LEDhookFunction"]=LEDhookFunction;
		NK2.Controls[key]["LEDhookGroup"]=LEDhookGroup;
		

		if (LEDstateType=="on"){//LED is always on
			state=1;
			NK2.Controls[key]["LEDstate"]=1;
			}
		else if (LEDstateType=="off"){//LED is always off
			state=0;
			NK2.Controls[key]["LEDstate"]=0;
			}
		else if (LEDstateType=="hook"){//LED is hooked to a mixxx control
			//engine.connectControl(NK2.Controls[key]["LEDhookGroup"], NK2.Controls[key]["LEDhookControl"], NK2.Controls[key]["LEDhookFunction"]);//set hook
			//engine.trigger(NK2.Controls[key]["LEDhookGroup"], NK2.Controls[key]["LEDhookControl"]);
			
			NK2.setHook(LEDhookGroup, LEDhookControl, LEDhookFunction);//set hook
			engine.trigger(LEDhookGroup, LEDhookControl);
			
			//evalstr=NK2.Controls[key]["LEDhookFunction"]+"("+engine.getValue(NK2.Controls[key]["LEDhookGroup"], NK2.Controls[key]["LEDhookControl"])+", '"+NK2.Controls[key]["LEDhookGroup"]+"', '"+NK2.Controls[key]["LEDhookControl"]+"');";
			//eval(evalstr);
			
			continue;
			}
		else if (thecontrol["LEDstateType"]=="eval"){//LED state determined by evaluating a javascript statement
			if (NK2.debug>2){print("##updateLEDs Eval: "+thecontrol["LEDstateEval"])};
			if (NK2.debug>2){print("##updateLEDs midino: "+midino)};
			if (NK2.debug>2){print("##evalstate: "+eval(thecontrol["LEDstateEval"]))};
			
			state=eval(LEDstateEval);
			NK2.Controls[key]["LEDstate"]=(state>0)?1:0;
			}//end if
			
		midi.sendShortMsg(NK2.midiChannel, midino, state*127);
		state=0;//reset state
		}//end for
	print("----------------UPDATE LEDS-----cycle state: "+NK2.cycleState);
	print("##############END LEDS######");
	};

NK2.indicatorLEDs = function indicatorLEDs(value,group,control){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	engine.stopTimer(NK2.flashTimer);//kill previous timer if exists
		
	NK2.flashon=0;
	NK2.flashcount=0;
	
	NK2.flashTimer=engine.beginTimer(100, "NK2.flashIndicators()");
	return true;
};

NK2.flashIndicators= function flashIndicators(){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};

	if (NK2.flashon==0){
		if (NK2.flashcount>=2){NK2.stopFlashTimer();return;}
		if (NK2.LEDBankIndicator!=false){midi.sendShortMsg(NK2.midiChannel, NK2.LEDBankIndicator, 0);}
		NK2.flashon=1;
		NK2.flashcount++;
		}else {
		if (NK2.flashcount>=2){NK2.stopFlashTimer();return;}
		if (NK2.LEDBankIndicator!=false){midi.sendShortMsg(NK2.midiChannel, NK2.LEDBankIndicator, 127);}
		NK2.flashon=0;
		NK2.flashcount++;
		}
	}

NK2.stopFlashTimer = function stopFlashTimer() {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	engine.stopTimer(NK2.flashTimer);
	NK2.flashcount=0;
	if (NK2.LEDBankIndicator!=false){midi.sendShortMsg(NK2.midiChannel, NK2.LEDBankIndicator, NK2.Controls[NK2.LEDBankIndicator]["LEDstate"]);};//restore state
	}

NK2.binaryHookLEDs = function lightButtonLEDs(value,group,control){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	//light button leds for simple binary values
	value=value*127;
	value=(value>127)?127:value;
	value=(value<0)?0:value;
	var state=(value>0)?1:0;

	//iterate through Controls to find button(s) that's hooked to this control
	for (var key in NK2.Controls){
		if (NK2.Controls[key]["LEDhookControl"]==control && NK2.Controls[key]["LEDhookGroup"]==group){
			var midino=key;
			midi.sendShortMsg(NK2.midiChannel, midino, value);
			NK2.Controls[key]["LEDstate"]=state;
		}
	}//end for
};

NK2.muteLEDs = function muteLEDs(value,group,control){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	//light deck mute leds, also cancels mute if volume changes

	//iterate through Controls to find button(s) that's hooked to this control
	for (var key in NK2.Controls){
		if (NK2.Controls[key]["LEDhookControl"]==control && NK2.Controls[key]["LEDhookGroup"]==group){
			var midino=key;
			var state=(value>0 || NK2.deckMuteVol[group]==-1)?0:1;
			midi.sendShortMsg(NK2.midiChannel, midino, state*127);
			NK2.Controls[key]["LEDstate"]=state;
			if(NK2.deckMuteVol[group]!=-1 && value>0){NK2.deckMuteVol[group]=-1;}
		}
	}//end for
};

NK2.vuMeterL = function vuMeterL(value,group,control){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	//light S buttons left to right for left VU meter

	value=(value*7)+1;
	for (var i=1; i<9; i++){
		if (value>=i){midi.sendShortMsg(NK2.midiChannel, NK2.Sbutton[i], 127);}else{midi.sendShortMsg(NK2.midiChannel, NK2.Sbutton[i], 0);}
	}
};

NK2.vuMeterR = function vuMeterR(value,group,control){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	//light S buttons left to right for left VU meter

	value=(value*7)+1;
	for (var i=1; i<9; i++){
		if (value>=i){midi.sendShortMsg(NK2.midiChannel, NK2.Mbutton[i], 127);}else{midi.sendShortMsg(NK2.midiChannel, NK2.Mbutton[i], 0);}
	}
};

NK2.beatIndicatorRow = function beatIndicatorRow(value,group,control){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	//light S buttons left to right for left VU meter

	value=(value*127);
	for (var i=1; i<9; i++){
		midi.sendShortMsg(NK2.midiChannel, NK2.Rbutton[i], value);
	}
};



//############################################################################
//BASIC FUNCTIONS - stuff that makes the other stuff work
//############################################################################
NK2.getFunctionName = function getFunctionName(){//return name of calling function
	var re = /function (.*?)\(/
    var s = NK2.getFunctionName.caller.toString();
    var m = re.exec( s )
    return m[1];
};

NK2.doNothing = function doNothing(){//dummy function - do nothing
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
    return false;
};

NK2.setHook = function setHook(group, control, func){//clear hook
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	var done=false;
	do{
		print("#####setting hook------------------------------>>>");
		print("##group: "+group);
		print("##control: "+control);
		print("##func: "+func);
		print("##doneset: "+done);
		done=engine.connectControl(group, control, func);
		print("##donesetpost: "+done);
	}while (done===false);
    return true;
};

NK2.clearHook = function clearHook(group, control, func){//clear hook
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	var done=false;
	do{
		print("#####clearing hook@@@@@@@@@@@@@@@@@@@@@@@");
		print("##group: "+group);
		print("##control: "+control);
		print("##func: "+func);
		print("##doneclear: "+done);
		done=engine.connectControl(group, control, func, true);
		print("##doneclearpost: "+done);
	}while (done===false);
    return true;
};

NK2.getControl = function getControl(control){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	
	var theControl=NK2.Controls[control][NK2.curNSMR][NK2.getCurModCode()];//get control for current mod settings
	if (theControl['isset']===false || typeof theControl === 'undefined'){
		theControl=NK2.Controls[control][NK2.curNSMR][NK2.bin2dec(NK2.cycleState.toString()+NK2.bankSelectState.toString()+"00")];//get control without mods, only check cycle state and bank select
		if (theControl['isset']===false || typeof theControl === 'undefined'){
			theControl=NK2.Controls[control][NK2.curNSMR][NK2.bin2dec("0"+NK2.bankSelectState.toString()+"00")];//get control without mods or cycle - ONLY bank select
			if (theControl['isset']===false || typeof theControl === 'undefined'){
				theControl=NK2.Controls[control][NK2.curNSMR][0];//get control without mods, cycle or bank select
			};

		};
	};
	
	if (theControl['isset']===false || typeof theControl === 'undefined'){return false;}
	
	//returns the array of control commands and LED state functions for the control pressed for the current bank/mods
	return theControl;
};
	
NK2.getCurModCode = function getCurModCode(control){//returns the current mod code value
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	
	var codestr=NK2.cycleState.toString()+NK2.bankSelectState.toString()+NK2.MOD1state.toString()+NK2.MOD2state.toString();
	//print("#####modcode: "+codestr);
	return parseInt(codestr, 2);
};

NK2.bin2dec = function bin2dec(str){//convert binary string to decimal
	return parseInt(str, 2);
};

//############################################################################
//FUNCTIONALITY - stuff that does stuff
//############################################################################

NK2.knobAdjust = function knobAdjust(group, control, value, min, max){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	if (group=="default"){group=NK2.Deck[NK2.curDeck];};
	//use a knob to adjust a mixxx control
	
	var range=max-min;
	var newValue=min+((value/127)*range);
	print("#########value: "+value);
	print("#########newval: "+newValue);
	if (newValue>max)newValue=max;
	if (newValue<min)newValue=min;
	engine.setValue(group, control, newValue);
	}
	
NK2.logKnobAdjust = function logKnobAdjust(group, control, value, min, max){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	if (group=="default"){group=NK2.Deck[NK2.curDeck];};
	//use a knob to adjust a mixxx control
	//on a logarithmic scale (0 min, 4 max)
	//K = 4*(V)^2
	//V = (K/4)^0.5 -- K= value coming from prog (0-4, logarithmic scale), V=value adjusted to linear 0-1 
	var range=max-min;
	var adjustedValue=value/127;//adjust range so it's on a scale of 0 to 1 - logarithmic
	var newValue=min+(range*Math.pow((adjustedValue),2));

	if (newValue>max)newValue=max;
	if (newValue<min)newValue=min;
	engine.setValue(group, control, newValue);
	}

NK2.wavezoomAll = function wavezoomAll(value){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
		
	var range=6-1;
	var newValue=Math.round(1+((value/127)*range));
	if (newValue>6)newValue=6;
	if (newValue<1)newValue=1;
	if (NK2.lastwavevalue!=value){
		for (var i=1; i<9; i++){
			engine.setValue(NK2.Deck[i], "waveform_zoom", newValue);
		};
	}
	NK2.lastwavevalue=value;
	}

NK2.wavezoomDeck = function wavezoomDeck(value, group){
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	if (group=="default"){group=NK2.Deck[NK2.curDeck];};
	
	var range=6-1;
	var newValue=Math.round(1+((value/127)*range));
	if (newValue>6)newValue=6;
	if (newValue<1)newValue=1;
	if (NK2.lastwavevalue!=value){
		engine.setValue(group, "waveform_zoom", newValue);
	}
	NK2.lastwavevalue=value;
	}

NK2.mutePress = function mutePress(deck){//press mute button - toggle mute for a deck
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	
	if (NK2.deckMuteVol[deck]<0){//deck is currently not muted.  enable mute
	print("deckvol: "+NK2.deckMuteVol[deck]);
		NK2.deckMuteVol[deck]=engine.getValue(deck, "volume");
		engine.setValue(deck, "volume", 0);
	} else {//deck is currently muted.  Unmute
		engine.setValue(deck, "volume", NK2.deckMuteVol[deck]);
		NK2.deckMuteVol[deck]=-1;
	}
};

NK2.muteRelease = function muteRelease(deck){//release mute button
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	if(NK2.deckMuteVol[deck]!=-1){
		engine.setValue(deck, "volume", NK2.deckMuteVol[deck]);
		NK2.deckMuteVol[deck]=-1;
	};
	return true;
};

NK2.binControlPress = function binControlPress(deck, control){//press button - toggle binary control for a deck
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	if (deck=="default"){deck=NK2.Deck[NK2.curDeck];};
	
	var curstate=engine.getValue(deck, control);
	if (curstate===0){//control currently off
		engine.setValue(deck, control,1);
	}else{
		engine.setValue(deck, control,0);
	}
};

NK2.binControlRelease = function binControlMomentary(deck, control){//release momentary button - reset control to 0 on release
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	if (deck=="default"){deck=NK2.Deck[NK2.curDeck];};
	
	engine.setValue(deck, control,0);
};

NK2.releaseSetFalse = function releaseSetFalse(deck, control){//release pfl button
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	engine.setValue(deck, control,false);
	return true;
};

NK2.cuePress = function cuePress(cue){//press hotcue
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	if (NK2.debug>2){print("##cue: "+cue)};
	if (NK2.debug>2){print("##deck: "+NK2.Deck[NK2.curDeck])};
	engine.setValue(NK2.Deck[NK2.curDeck], "hotcue_"+cue+"_activate", 1);
	engine.setValue(NK2.Deck[NK2.curDeck], "hotcue_"+cue+"_activate", 0);
}

NK2.cueClear = function cueClear(cue){//clear hotcue
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	if (NK2.debug>2){print("##cue: "+cue)};
	if (NK2.debug>2){print("##deck: "+NK2.Deck[NK2.curDeck])};
	engine.setValue(NK2.Deck[NK2.curDeck], "hotcue_"+cue+"_clear", 1);
	engine.setValue(NK2.Deck[NK2.curDeck], "hotcue_"+cue+"_clear", 0);
}

NK2.beatjump = function (jumplen) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	//jumps back certain number of beats depending on knob modifier
	var curpos = engine.getValue(NK2.Deck[NK2.curDeck], "playposition")*engine.getValue(NK2.Deck[NK2.curDeck], "track_samples");
	var numbeats = jumplen;
	var backseconds = numbeats*(1/(engine.getValue(NK2.Deck[NK2.curDeck], "bpm")/60));
	var backsamples = backseconds*engine.getValue(NK2.Deck[NK2.curDeck], "track_samplerate")*2;//*2 to compensate for stereo samples
	var newpos = curpos-(backsamples);
	
	if (NK2.debug){print("backseconds: "+backseconds);}
	if (NK2.debug){print("backsamples: "+backsamples);}
	if (NK2.debug){print("curpos: "+curpos);}
	if (NK2.debug){print("newpos: "+newpos);}
	if (NK2.debug){print("numbeats: "+numbeats);}
	
	engine.setValue(NK2.Deck[NK2.curDeck], "playposition", newpos/engine.getValue(NK2.Deck[NK2.curDeck], "track_samples"));
	};

NK2.beatloop = function beatloop(looplen, value) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};

	//activate beatloop
	if (value>0){//button was pressed
		NK2.loopbuttonDown=true;
		engine.setValue(NK2.Deck[NK2.curDeck], "beatloop_"+looplen+"_activate", 1);
		} else {//button was released
		NK2.loopbuttonDown=false;
		};

	};
	
NK2.beatlooproll = function beatlooproll(looplen, value) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};

	//activate beatloop
	if (value>0){//button was pressed
		NK2.loopbuttonDown=true;
		engine.setValue(NK2.Deck[NK2.curDeck], "beatloop_"+looplen+"_activate", 1);
		} else {//button was released
		NK2.loopbuttonDown=false;
		};

	};

NK2.reloopButton = function reloopButton(value) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	
	if (value>0){//button was pressed
		engine.stopTimer(NK2.reloopTimer);
		NK2.loopbuttonDown=true;
		NK2.doreloop=true;
		NK2.reloopTimer = engine.beginTimer(500, "NK2.disablereloop()", true);
		} else {//button was released
		NK2.loopbuttonDown=false;
		if (NK2.doreloop===true) {engine.setValue(NK2.Deck[NK2.curDeck], "reloop_exit", 1);engine.setValue(NK2.Deck[NK2.curDeck], "reloop_exit", 0);};
		NK2.doreloop=true;
		engine.stopTimer(NK2.reloopTimer);
		};

	};

NK2.loopin = function loopin(value) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};

	//activate beatloop
	if (value>0){//button was pressed
		engine.stopTimer(NK2.loopinTimer);
		NK2.loopinbuttonDown=true;
		NK2.doloopin=true;
		NK2.loopinTimer = engine.beginTimer(500, "NK2.disableloopin()", true);
		} else {//button was released
		NK2.loopinbuttonDown=false;
		if (NK2.doloopin===true) {engine.setValue(NK2.Deck[NK2.curDeck], "loop_in", 1);engine.setValue(NK2.Deck[NK2.curDeck], "loop_in", 0);};
		NK2.doloopin=true;
		engine.stopTimer(NK2.loopinTimer);
		};

	};
	
NK2.loopout = function loopout(value) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};

	//activate beatloop
	if (value>0){//button was pressed
		engine.stopTimer(NK2.loopoutTimer);
		NK2.loopoutbuttonDown=true;
		NK2.doloopout=true;
		NK2.loopoutTimer = engine.beginTimer(500, "NK2.disableloopout()", true);
		} else {//button was released
		NK2.loopoutbuttonDown=false;
		if (NK2.doloopout===true) {engine.setValue(NK2.Deck[NK2.curDeck], "loop_out", 1);engine.setValue(NK2.Deck[NK2.curDeck], "loop_out", 0);};
		NK2.doloopout=true;
		engine.stopTimer(NK2.loopoutTimer);
		};

	};
	
NK2.disablereloop = function disablereloop() {
	//timed function - fires half a second after pressing reloop.  Don't do the reloop if you hold down the button (so you can move the loop without exiting)
	NK2.doreloop=false;
	};

NK2.disableloopin = function disableloopin() {
	//timed function - fires half a second after pressing reloop.  Don't do the reloop if you hold down the button (so you can move the loop without exiting)
	NK2.doloopin=false;
	};

NK2.disableloopout = function disableloopout() {
	//timed function - fires half a second after pressing reloop.  Don't do the reloop if you hold down the button (so you can move the loop without exiting)
	NK2.doloopout=false;
	};

NK2.loopminus = function loopminus() {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	//shrinks loop or moves loop back
	if (NK2.loopbuttonDown !== true && NK2.loopinbuttonDown!==true && NK2.loopoutbuttonDown!==true){engine.setValue(NK2.Deck[NK2.curDeck], "loop_halve", 1); engine.setValue(NK2.Deck[NK2.curDeck], "loop_halve", 0); return false;}//shrink loop if no loop button down 
	else if (NK2.loopbuttonDown === true && engine.getValue(NK2.Deck[NK2.curDeck], "loop_start_position")>=0 && engine.getValue(NK2.Deck[NK2.curDeck], "loop_end_position")>=0 ){
		//move loop
		var interval =	NK2.loopmove*engine.getValue(NK2.Deck[NK2.curDeck], "track_samples")/engine.getValue(NK2.Deck[NK2.curDeck], "duration");
		var start = engine.getValue(NK2.Deck[NK2.curDeck], "loop_start_position");
		var end = engine.getValue(NK2.Deck[NK2.curDeck], "loop_end_position");
		engine.setValue(NK2.Deck[NK2.curDeck], "loop_start_position", start-interval);
		engine.setValue(NK2.Deck[NK2.curDeck], "loop_end_position", end-interval);
		return true;
		}
	else if (NK2.loopinbuttonDown === true && engine.getValue(NK2.Deck[NK2.curDeck], "loop_start_position")>=0){
		//move loop in point
		var interval =	NK2.loopmove*engine.getValue(NK2.Deck[NK2.curDeck], "track_samples")/engine.getValue(NK2.Deck[NK2.curDeck], "duration");
		var start = engine.getValue(NK2.Deck[NK2.curDeck], "loop_start_position");
		engine.setValue(NK2.Deck[NK2.curDeck], "loop_start_position", start-interval);
		return true;
		}
	else if (NK2.loopoutbuttonDown === true && engine.getValue(NK2.Deck[NK2.curDeck], "loop_end_position")>=0){
		//move loop out point
		var interval =	NK2.loopmove*engine.getValue(NK2.Deck[NK2.curDeck], "track_samples")/engine.getValue(NK2.Deck[NK2.curDeck], "duration");
		var end = engine.getValue(NK2.Deck[NK2.curDeck], "loop_end_position");
		engine.setValue(NK2.Deck[NK2.curDeck], "loop_end_position", end-interval);
		return true;
		};
	};

NK2.loopplus = function loopplus() {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	//grows loop or moves loop forward
	if (NK2.loopbuttonDown !== true && NK2.loopinbuttonDown!==true && NK2.loopoutbuttonDown!==true){engine.setValue(NK2.Deck[NK2.curDeck], "loop_double", 1); engine.setValue(NK2.Deck[NK2.curDeck], "loop_double", 0); return false;}//shrink loop if no loop button down 
	else if (NK2.loopbuttonDown === true && engine.getValue(NK2.Deck[NK2.curDeck], "loop_start_position")>=0 && engine.getValue(NK2.Deck[NK2.curDeck], "loop_end_position")>=0 ){
		//move loop
		var interval =	NK2.loopmove*engine.getValue(NK2.Deck[NK2.curDeck], "track_samples")/engine.getValue(NK2.Deck[NK2.curDeck], "duration");
		var start = engine.getValue(NK2.Deck[NK2.curDeck], "loop_start_position");
		var end = engine.getValue(NK2.Deck[NK2.curDeck], "loop_end_position");
		engine.setValue(NK2.Deck[NK2.curDeck], "loop_start_position", start+interval);
		engine.setValue(NK2.Deck[NK2.curDeck], "loop_end_position", end+interval);
		return true;
		}
	else if (NK2.loopinbuttonDown === true && engine.getValue(NK2.Deck[NK2.curDeck], "loop_start_position")>=0){
		//move loop in point
		var interval =	NK2.loopmove*engine.getValue(NK2.Deck[NK2.curDeck], "track_samples")/engine.getValue(NK2.Deck[NK2.curDeck], "duration");
		var start = engine.getValue(NK2.Deck[NK2.curDeck], "loop_start_position");
		engine.setValue(NK2.Deck[NK2.curDeck], "loop_start_position", start+interval);
		return true;
		}
	else if (NK2.loopoutbuttonDown === true && engine.getValue(NK2.Deck[NK2.curDeck], "loop_end_position")>=0){
		//move loop out point
		var interval =	NK2.loopmove*engine.getValue(NK2.Deck[NK2.curDeck], "track_samples")/engine.getValue(NK2.Deck[NK2.curDeck], "duration");
		var end = engine.getValue(NK2.Deck[NK2.curDeck], "loop_end_position");
		engine.setValue(NK2.Deck[NK2.curDeck], "loop_end_position", end+interval);
		return true;
		};
	};

//############################################################################
//TEST FUNCTIONS - stuff to check out if the other stuff works
//############################################################################
NK2.test = function test(channel, control, value, status, group) {
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	
	print("channel: "+channel);
	print("control: "+control.toString(16));
	print("value: "+value);
	print("status: "+status);
	print("group: "+group);
	print("test: ##"+status.toString(16)+control.toString(16)+"##");
	print("loopbuttondown: ##"+NK2.loopbuttonDown+"##");
	print("NK2.looplen: ##"+NK2.looplen+"##");
	print("NK2.looplen.tostring: ##"+NK2.looplen.toString(16)+"##");
	print("NK2.looptype: ##"+NK2.looptype+"##");
	print("test: ##"+2+"##");
};

NK2.testLED=0;//used by test function
NK2.testleds=function testleds(){//sends LED on messages to all
	midi.sendShortMsg(0xB0, NK2.testLED, 0);
	if (NK2.testLED>=127){return false; } 
	NK2.testLED++
	midi.sendShortMsg(0xB0, NK2.testLED, 0x7F);	
	NK2.LEDtimer=engine.beginTimer(100, "NK2.testleds()", true);
}
	
//############################################################################
//CONTROL CONFIG - initialize control command object and LED state tracker, bind commands to controls
//############################################################################

//initialize control object function - iterates through buttons and sets up empty slots for command arrays - unset controls are "false"
NK2.initControls = function initControls(obj) {//initialize controls array
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};
	
	for (var x in obj){
		NK2.Controls[obj[x]]={"LEDstate":0, "LEDstateType":false, "LEDstateEval":false, "LEDhookControl":false, "LEDhookFunction":false, "LEDhookGroup":false};//init with empty LED tracking properties
		
		for (var y in NK2.Banks){//iterate through bank modes
			NK2.Controls[obj[x]][NK2.Banks[y]]={};//init empty
			
			for (var z=0;z<16;z++){//iterate through mod codes
				NK2.Controls[obj[x]][NK2.Banks[y]][z]={"isset":false, "pressEval":false, "releaseEval":false, "LEDstateType":false, "LEDstateEval":false, "LEDhookControl":false, "LEDhookFunction":false, "LEDhookGroup":false};
				};//end for z
			};//end for y
		};//end for x
	};





NK2.setup = function setup(obj) {//initialize controls array
	if (NK2.debug>2){print("##function: "+NK2.getFunctionName())};

	
//initialize control object for buttons, knobs, etc.
NK2.Controls={};

NK2.initControls(NK2.Knob);
NK2.initControls(NK2.Sbutton);
NK2.initControls(NK2.Mbutton);
NK2.initControls(NK2.Rbutton);
NK2.initControls(NK2.leftButton);

//CONTROL SYNTAX
//NK2.Controls[midino]["N" (none), "S", "M", or "R"][MOD state: no mod = 0, bank select = 1, MOD 1 on = 2, MOD 2 on = 3]
//'PressEval', 'ReleaseEval' = code eval'd when button is pressed or released 

/* MOD STATE - binary based
 * 1st digit - cycle button - this is a toggle (others are momentary on), and has a LED (others don't)
 * 2nd digit - bank select (marker set) button
 * 3rd digit - MOD 1 - track down
 * 4th digit - MOD 2 - track up
 * 
 * eg:
 * 0100 - (dec. 4) - bank select only
 * 1101 - (dec. 5) - cycle, bank sel. and MOD 2
 * 
 * decimal equivalents - 3 buttons = 8 states - 4 buttons, 16 states
0000 - 0
0001 - 1
0010 - 2
0011 - 3
0100 - 4
0101 - 5
0110 - 6
0111 - 7
1000 - 8
1001 - 9
1010 - 10
1011 - 11
1100 - 12
1101 - 13
1110 - 14
1111 - 15
 */


//eval strs can be false if no function on press or release - eg: knobs are false on release, since they don't release
//LEDstateType = false if control has no LED, "hook" if using connectControl, "eval" if using an eval statement to get initial LED state
//LEDstateType = "off" if LED is always off, "on" if LED is always on
//for HOOKS: LEDhookControl = mixxx control to connect to, LEDhookFunction = script function
//for HOOKS: LEDhookGroup = mixxx control group - literals accepted (eg: "[Channel1]", or "default", "sync" or "alt" = default, sync or alt deck for that channel
//for EVAL: LEDstateEval = javascript to evaluate to get LED state
//init: {"isset":false, "pressEval":false, "releaseEval":false, "LEDstateType":false, "LEDstateEval":false, "LEDhookControl":false, "LEDhookFunction":false, "LEDhookGroup":false}\

//bank select and mod buttons - marker set down and up
for (var j=0;j<4;j++){
	NK2.Controls[NK2.leftButton["mset"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.modpress("BANK");', "releaseEval":'NK2.modrelease("BANK");', "LEDstateType":"on"};
	NK2.Controls[NK2.leftButton["mdown"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.modpress("MOD1");', "releaseEval":'NK2.modrelease("MOD1");', "LEDstateType":"on"};
	NK2.Controls[NK2.leftButton["mup"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.modpress("MOD2");', "releaseEval":'NK2.modrelease("MOD2");', "LEDstateType":"on"};
	NK2.Controls[NK2.leftButton["cycle"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.cyclepress();', "releaseEval":false, "LEDstateType":"eval", "LEDstateEval":'(NK2.cycleState==0)?0:1;'};
};

//bank select button pressed - initialize bank select buttons
for (var i=1;i<9;i++){//i = strip number
	for (var j=0;j<4;j++){
		NK2.Controls[NK2.Sbutton[i]][NK2.Banks[j]][NK2.MODcodes["0100"]]={"isset":true, "pressEval":'NK2.bankSelect('+i+', "S");', "LEDstateType":"on"};
		NK2.Controls[NK2.Mbutton[i]][NK2.Banks[j]][NK2.MODcodes["0100"]]={"isset":true, "pressEval":'NK2.bankSelect('+i+', "M");', "LEDstateType":"on"};
		NK2.Controls[NK2.Rbutton[i]][NK2.Banks[j]][NK2.MODcodes["0100"]]={"isset":true, "pressEval":'NK2.bankSelect('+i+', "R");', "LEDstateType":"on"};
		};
	};

//set common left button controls for S, M, and R modes (these controls can be overridden by resetting them later in the script) 	
for (var j=1;j<4;j++){
	NK2.Controls[NK2.leftButton["rev"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.binControlPress("default", "quantize");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'quantize', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.leftButton["ff"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.binControlPress("default", "keylock");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'keylock', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.leftButton["stop"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.binControlPress("default", "beats_translate_curpos");', "releaseEval":'NK2.binControlRelease("default", "beats_translate_curpos");', "LEDstateType":"off"};
	NK2.Controls[NK2.leftButton["play"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.binControlPress("default", "play");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'play', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.leftButton["rec"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.binControlPress("default", "cue_default");', "releaseEval":'NK2.binControlRelease("default", "cue_default");', "LEDstateType":"hook", "LEDhookControl":'cue_default', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.leftButton["trdown"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.binControlPress("default", "back");', "releaseEval":'NK2.binControlRelease("default", "back");', "LEDstateType":"off"};
	NK2.Controls[NK2.leftButton["trup"]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.binControlPress("default", "fwd");', "releaseEval":'NK2.binControlRelease("default", "fwd");', "LEDstateType":"off"};
	};
	
//NO BANK/TRACK SELECTED - bank select button off - set sfx, mute and pfl buttons
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["N"][0]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "flanger");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'flanger', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	NK2.Controls[NK2.Mbutton[i]]["N"][0]={"isset":true, "pressEval":'NK2.mutePress("'+NK2.Deck[i]+'");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'volume', "LEDhookFunction":'NK2.muteLEDs', "LEDhookGroup":NK2.Deck[i]};
	NK2.Controls[NK2.Rbutton[i]]["N"][0]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "pfl");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'pfl', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	};

//CYCLE ON - momentary sfx, and mute buttons
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["N"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "flanger");', "releaseEval":'NK2.binControlRelease("'+NK2.Deck[i]+'", "flanger");', "LEDstateType":"hook", "LEDhookControl":'flanger', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	NK2.Controls[NK2.Mbutton[i]]["N"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":'NK2.mutePress("'+NK2.Deck[i]+'");', "releaseEval":'NK2.muteRelease("'+NK2.Deck[i]+'");', "LEDstateType":"hook", "LEDhookControl":'volume', "LEDhookFunction":'NK2.muteLEDs', "LEDhookGroup":NK2.Deck[i]};
	};

//N mode + MOD1 - lo mid hi kills
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["N"][2]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "filterHighKill");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'filterHighKill', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	NK2.Controls[NK2.Mbutton[i]]["N"][2]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "filterMidKill");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'filterMidKill', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	NK2.Controls[NK2.Rbutton[i]]["N"][2]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "filterLowKill");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'filterLowKill', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	};

//CYCLE ON - N mode + MOD1 - MOMENTARY lo mid hi kills
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["N"][NK2.MODcodes["1010"]]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "filterHighKill");', "releaseEval":'NK2.binControlRelease("'+NK2.Deck[i]+'", "filterHighKill");', "LEDstateType":"hook", "LEDhookControl":'filterHighKill', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	NK2.Controls[NK2.Mbutton[i]]["N"][NK2.MODcodes["1010"]]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "filterMidKill");', "releaseEval":'NK2.binControlRelease("'+NK2.Deck[i]+'", "filterMidKill");', "LEDstateType":"hook", "LEDhookControl":'filterMidKill', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	NK2.Controls[NK2.Rbutton[i]]["N"][NK2.MODcodes["1010"]]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "filterLowKill");', "releaseEval":'NK2.binControlRelease("'+NK2.Deck[i]+'", "filterLowKill");', "LEDstateType":"hook", "LEDhookControl":'filterLowKill', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	};

//N mode + MOD2 - keylock quantize
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["N"][1]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "quantize");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'quantize', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	NK2.Controls[NK2.Mbutton[i]]["N"][1]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "keylock");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'keylock', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	NK2.Controls[NK2.Rbutton[i]]["N"][1]={"isset":true, "pressEval":'NK2.binControlPress("'+NK2.Deck[i]+'", "beatsync");', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'beatsync', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":NK2.Deck[i]};
	};

//M mode - hotcues 1-24
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["M"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.cuePress('+i+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+i+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.Mbutton[i]]["M"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.cuePress('+(i+8)+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+(i+8)+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.Rbutton[i]]["M"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.cuePress('+(i+16)+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+(i+16)+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	};

//M mode, with cycle toggled on - hotcues 9-32
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["M"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":'NK2.cuePress('+(i+8)+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+(i+8)+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.Mbutton[i]]["M"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":'NK2.cuePress('+(i+16)+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+(i+16)+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.Rbutton[i]]["M"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":'NK2.cuePress('+(i+24)+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+(i+24)+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	};

//M mode - CLEAR (with MOD2) hotcues 1-24
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["M"][NK2.MODcodes["0001"]]={"isset":true, "pressEval":'NK2.cueClear('+i+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+i+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.Mbutton[i]]["M"][NK2.MODcodes["0001"]]={"isset":true, "pressEval":'NK2.cueClear('+(i+8)+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+(i+8)+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.Rbutton[i]]["M"][NK2.MODcodes["0001"]]={"isset":true, "pressEval":'NK2.cueClear('+(i+16)+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+(i+16)+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	};

//M mode, with cycle toggled on - CLEAR (with MOD2) hotcues 9-32
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["M"][NK2.MODcodes["1001"]]={"isset":true, "pressEval":'NK2.cueClear('+(i+8)+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+(i+8)+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.Mbutton[i]]["M"][NK2.MODcodes["1001"]]={"isset":true, "pressEval":'NK2.cueClear('+(i+16)+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+(i+16)+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.Rbutton[i]]["M"][NK2.MODcodes["1001"]]={"isset":true, "pressEval":'NK2.cueClear('+(i+24)+');', "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'hotcue_'+(i+24)+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	};

//R mode with cycle - loops and beatjumps
var looplens={1:.125, 2:.25, 3:.5, 4:1, 5:2, 6:4, 7:8, 8:16}//array of loop and beatjump lengths
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["R"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.beatjump('+looplens[i]+');', "releaseEval":false, "LEDstateType":"off"};
	NK2.Controls[NK2.Mbutton[i]]["R"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.beatlooproll('+looplens[i]+', value);', "releaseEval":'NK2.beatlooproll('+looplens[i]+', value);', "LEDstateType":"hook", "LEDhookControl":'beatloop_'+looplens[i]+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.Rbutton[i]]["R"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.beatloop('+looplens[i]+', value);', "releaseEval":'NK2.beatloop('+looplens[i]+', value);', "LEDstateType":"hook", "LEDhookControl":'beatloop_'+looplens[i]+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	};
//loophalve loopdouble
NK2.Controls[NK2.leftButton["trdown"]]["R"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.loopminus();', "releaseEval":false, "LEDstateType":"off"};
NK2.Controls[NK2.leftButton["trup"]]["R"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.loopplus();', "releaseEval":false, "LEDstateType":"off"};
//reloop exit
NK2.Controls[NK2.leftButton["stop"]]["R"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.reloopButton(value);', "releaseEval":'NK2.reloopButton(value);', "LEDstateType":"hook", "LEDhookControl":'loop_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
//loopin loopout
NK2.Controls[NK2.leftButton["rev"]]["R"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.loopin(value);', "releaseEval":'NK2.loopin(value);', "LEDstateType":"hook", "LEDhookControl":'loop_start_position', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
NK2.Controls[NK2.leftButton["ff"]]["R"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.loopout(value);', "releaseEval":'NK2.loopout(value);', "LEDstateType":"hook", "LEDhookControl":'loop_end_position', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};

//R mode with cycle - shorter loops and beatjumps
var shortlooplens={1:0.03125, 2:0.0625, 3:0.125, 4:0.25, 5:0.5, 6:1, 7:2, 8:4};//array of shorter loop and beatjump lengths
for (var i=1;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["R"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":'NK2.beatjump('+shortlooplens[i]+');', "releaseEval":false, "LEDstateType":"off"};
	NK2.Controls[NK2.Mbutton[i]]["R"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":'NK2.beatlooproll('+shortlooplens[i]+', value);', "releaseEval":'NK2.beatlooproll('+shortlooplens[i]+', value);', "LEDstateType":"hook", "LEDhookControl":'beatloop_'+shortlooplens[i]+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	NK2.Controls[NK2.Rbutton[i]]["R"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":'NK2.beatloop('+shortlooplens[i]+', value);', "releaseEval":'NK2.beatloop('+shortlooplens[i]+', value);', "LEDstateType":"hook", "LEDhookControl":'beatloop_'+shortlooplens[i]+'_enabled', "LEDhookFunction":'NK2.binaryHookLEDs', "LEDhookGroup":"default"};
	};

//S mode - vumeters and beat indicator
NK2.Controls[NK2.Sbutton[1]]["S"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'VuMeterL', "LEDhookFunction":'NK2.vuMeterL', "LEDhookGroup":"default"};
NK2.Controls[NK2.Mbutton[1]]["S"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'VuMeterR', "LEDhookFunction":'NK2.vuMeterR', "LEDhookGroup":"default"};
NK2.Controls[NK2.Rbutton[1]]["S"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'beat_active', "LEDhookFunction":'NK2.beatIndicatorRow', "LEDhookGroup":"default"};
for (var i=2;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["S"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"off"};
	NK2.Controls[NK2.Mbutton[i]]["S"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"off"};
	NK2.Controls[NK2.Rbutton[i]]["S"][NK2.MODcodes["1000"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"off"};
	};

//master VU meter - MOD1 and MOD2 - N mode
NK2.Controls[NK2.Sbutton[1]]["N"][NK2.MODcodes["0011"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'VuMeterL', "LEDhookFunction":'NK2.vuMeterL', "LEDhookGroup":"[Master]"};
NK2.Controls[NK2.Mbutton[1]]["N"][NK2.MODcodes["0011"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"hook", "LEDhookControl":'VuMeterR', "LEDhookFunction":'NK2.vuMeterR', "LEDhookGroup":"[Master]"};
NK2.Controls[NK2.Rbutton[1]]["N"][NK2.MODcodes["0011"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"off"};
for (var i=2;i<9;i++){
	NK2.Controls[NK2.Sbutton[i]]["N"][NK2.MODcodes["0011"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"off"};
	NK2.Controls[NK2.Mbutton[i]]["N"][NK2.MODcodes["0011"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"off"};
	NK2.Controls[NK2.Rbutton[i]]["N"][NK2.MODcodes["0011"]]={"isset":true, "pressEval":false, "releaseEval":false, "LEDstateType":"off"};
	};

//############# KNOBS
//N mode
NK2.Controls[NK2.Knob[1]]["N"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.wavezoomAll(value);', "LEDstateType":"off"};
NK2.Controls[NK2.Knob[2]]["N"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.logKnobAdjust("[Master]", "headVolume", value, 0, 5);', "LEDstateType":"off"};
NK2.Controls[NK2.Knob[3]]["N"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.knobAdjust("[Master]", "headMix", value, -1, 1);', "LEDstateType":"off"};
NK2.Controls[NK2.Knob[4]]["N"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.knobAdjust("[Master]", "balance", value, -1, 1);', "LEDstateType":"off"};
NK2.Controls[NK2.Knob[5]]["N"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.logKnobAdjust("[Master]", "volume", value, 0, 5);', "LEDstateType":"off"};
NK2.Controls[NK2.Knob[6]]["N"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.knobAdjust("[Flanger]", "lfoPeriod", value, 50000, 2000000);', "LEDstateType":"off"};
NK2.Controls[NK2.Knob[7]]["N"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.knobAdjust("[Flanger]", "lfoDepth", value, 0, 1);', "LEDstateType":"off"};
NK2.Controls[NK2.Knob[8]]["N"][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.knobAdjust("[Flanger]", "lfoDelay", value, 50, 10000);', "LEDstateType":"off"};

//SMR MODES
for (var j=1;j<4;j++){
	NK2.Controls[NK2.Knob[1]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.wavezoomDeck(value, "default");', "LEDstateType":"off"};
	NK2.Controls[NK2.Knob[2]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.logKnobAdjust("default", "filterLow", value, 0, 4);', "LEDstateType":"off"};
	NK2.Controls[NK2.Knob[3]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.logKnobAdjust("default", "filterMid", value, 0, 4);', "LEDstateType":"off"};
	NK2.Controls[NK2.Knob[4]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.logKnobAdjust("default", "filterHigh", value, 0, 4);', "LEDstateType":"off"};
	NK2.Controls[NK2.Knob[5]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.logKnobAdjust("default", "pregain", value, 0, 4);', "LEDstateType":"off"};
	NK2.Controls[NK2.Knob[6]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.knobAdjust("default", "rate", value, -1, 1);', "LEDstateType":"off"};
	//NK2.Controls[NK2.Knob[7]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.knobAdjust("default", "");', "LEDstateType":"off"};
	//NK2.Controls[NK2.Knob[8]][NK2.Banks[j]][NK2.MODcodes["0000"]]={"isset":true, "pressEval":'NK2.knobAdjust("default", "");', "LEDstateType":"off"};
}

};//end setup function 


/*
NOTES:
* 
*/
